
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

int seed = 27182;
int rand_(int x)
{
  x += seed;
  x = (x << 13) ^ x;
  return (x * (x * x * 15731 + 789221) + 1376312589) & 0x7fffffff;
}

double interpolation(int v1, int v2, double x)
{
  double temp = (cos(x * M_PI) + 1.0) * 0.5;
  return temp * v1 + (1.0 - temp) * v2;
}

double perlin(int x, int y, int lh, int lv, int a)
{
  int r1, r2, r3, r4, temp;
  r1 = rand_((y / lv) * 10000 + (x / lh)) % a;
  r2 = rand_((y / lv) * 10000 + (x / lh) + 1) % a;
  r3 = rand_(((y / lv) + 1) * 10000 + (x / lh)) % a;
  r4 = rand_(((y / lv) + 1) * 10000 + (x / lh) + 1) % a;
  temp = interpolation(
    interpolation(r1, r2, (double)(x % lh) / lh),
    interpolation(r3, r4, (double)(x % lh) / lh),
    (double)(y % lv) / lv);
  return temp;
}

struct _inselfeld
{
  unsigned short bebauung;
  unsigned char x_pos;
  unsigned char y_pos;
  unsigned int rot : 2;
  unsigned int ani : 4;
  unsigned int a : 8;
  unsigned int spieler : 3;
  unsigned int zufall : 5;
  unsigned int spieler_ : 3;
  unsigned int c : 7;
};

typedef struct _inselfeld inselfeld_t;

char *insel5_id = "INSEL5";
struct _insel5_header
{
  unsigned char unknown_a;
  unsigned int unknown_b;
  unsigned int unknown_c;
  unsigned int inselhaus_offset;
  unsigned char unknown_d;
  unsigned char width;
  unsigned char height;
  unsigned char unknown_e;
  unsigned short x_pos;
  unsigned short y_pos;
  unsigned short unknown_f;
  unsigned short unknown_g;
  unsigned char bytes1[84];
  unsigned short base;
  unsigned short unknown_h;
  unsigned char sued;
  unsigned char diff;
  unsigned char bytes2[14];
} __attribute__((packed));

typedef struct _insel5_header insel5_header_t;

char *inselhaus_id = "INSELHAUS";
struct _inselhaus_header
{
  short a;
  int b;
  unsigned int size;
} __attribute__((packed));

typedef struct _inselhaus_header inselhaus_header_t;


#define ABS(a) (((a) >= 0) ? (a) : -(a))
#define MAX(a, b) (((a) > (b)) ? (a) : (b))
#define MIN(a, b) (((a) < (b)) ? (a) : (b))

void main(int argc, char **argv)
{
  unsigned char width = 70;
  unsigned char height = 60;
  
  inselfeld_t daten[70 * 60];
  double temp[70 * 60];
  int temp2[70 * 60];
  int temp3[70 * 60];
  
  /* Höhenkarte erzeugen */
  int x, y;
  for (y = 0; y < height; y++)
  {
    for (x = 0; x < width; x++)
    {
      temp[y * width + x] =
	((perlin(x, y, 20, 20, 100))
	* (cos (M_PI / 20.0 * MAX(0, MAX(ABS(x - 35), ABS(y - 30) + 5) - 15)) + 1.0) * 0.5
	+ perlin(x + 100, y, 10, 10, 70)
	+ perlin(x + 200, y, 5, 5, 50))
	* cos (M_PI_2 / 10.0 * MAX(0, MAX(ABS(x - 35), ABS(y - 30) + 5) - 25))
	- 75.0;
/*	- 40
	- MAX(0, MAX(ABS(x - 35), ABS(y - 30) + 5) - 25) * 10;*/
      if (temp[y * width + x] >= 0)
	temp2[y * width + x] = 1;
      else
	temp2[y * width + x] = 0;
    }
  }
  
  /* Median-Filter */
  for (y = 1; y < height - 1; y++)
  {
    for (x = 1; x < width - 1; x++)
    {
      if (temp2[(y - 1) * width + x - 1] + temp2[(y - 1) * width + x] + temp2[(y - 1) * width + x + 1]
	+ temp2[y * width + x - 1] + temp2[y * width + x] + temp2[y * width + x - 1]
	+ temp2[(y + 1) * width + x - 1] + temp2[(y + 1) * width + x] + temp2[(y + 1) * width + x + 1] <= 4)
	temp3[y * width + x] = 0;
      else
	temp3[y * width + x] = 1;
    }
  }
  
  /* Küste hinzufügen */
  for (y = 1; y < height - 1; y++)
  {
    for (x = 1; x < width - 1; x++)
    {
      if (temp3[y * width + x] == 0)
      {
	if ((temp3[(y - 1) * width + x - 1] == 1) || (temp3[(y - 1) * width + x] == 1) || (temp3[(y - 1) * width + x + 1] == 1)
	  || (temp3[y * width + x - 1] == 1) || (temp3[y * width + x] == 1) || (temp3[y * width + x + 1] == 1)
	  || (temp3[(y + 1) * width + x - 1] == 1) || (temp3[(y + 1) * width + x] == 1) || (temp3[(y + 1) * width + x + 1] == 1))
	  temp3[y * width + x] = 2;
      }
    }
  }
  
  /* Filtern */
  int active;
  do
  {
    active = 0;
    for (y = 1; y < height - 1; y++)
    {
      for (x = 1; x < width - 1; x++)
      {
	if (temp3[y * width + x] == 2)
	{
	  if (((temp3[(y - 1) * width + x] == 1) && (temp3[y * width + x - 1] == 1) && (temp3[(y + 1) * width + x] == 1))
	    || ((temp3[(y - 1) * width + x] == 1) && (temp3[y * width + x + 1] == 1) && (temp3[(y + 1) * width + x] == 1))
	    || ((temp3[(y - 1) * width + x] == 1) && (temp3[y * width + x - 1] == 1) && (temp3[y * width + x + 1] == 1))
	    || ((temp3[(y + 1) * width + x] == 1) && (temp3[y * width + x - 1] == 1) && (temp3[y * width + x + 1] == 1)))
	  {
	    temp3[y * width + x] = 1;
	    active = 1;
	  }
	}
      }
    }
  } while (active);
  do
  {
    active = 0;
    for (y = 1; y < height - 1; y++)
    {
      for (x = 1; x < width - 1; x++)
      {
	if (temp3[y * width + x] <= 0)
	{
	  if (((temp3[(y - 1) * width + x] >= 2) && (temp3[y * width + x - 1] >= 2) && (temp3[(y + 1) * width + x] >= 2))
	    || ((temp3[(y - 1) * width + x] >= 2) && (temp3[y * width + x + 1] >= 2) && (temp3[(y + 1) * width + x] >= 2))
	    || ((temp3[(y - 1) * width + x] >= 2) && (temp3[y * width + x - 1] >= 2) && (temp3[y * width + x + 1] >= 2))
	    || ((temp3[(y + 1) * width + x] >= 2) && (temp3[y * width + x - 1] >= 2) && (temp3[y * width + x + 1] >= 2)))
	  {
	    temp3[y * width + x] = 20;
	    active = 1;
	  }
	}
      }
    }
  } while (active);
  
  for (y = 1; y < height - 1; y++)
  {
    for (x = 0; x < width - 1; x++)
    {
      if (temp3[y * width + x] >= 2)
      {
	if ((temp3[(y - 1) * width + x] == 1))
	  temp3[y * width + x] = 2;
	if ((temp3[y * width + x + 1] == 1))
	  temp3[y * width + x] = 3;
	if ((temp3[(y + 1) * width + x] == 1))
	  temp3[y * width + x] = 4;
	if ((temp3[y * width + x - 1] == 1))
	  temp3[y * width + x] = 5;
	
	if ((temp3[(y + 1) * width + x] == 1) && (temp3[y * width + x - 1] == 1))
	  temp3[y * width + x] = 10;
	if ((temp3[(y - 1) * width + x] == 1) && (temp3[y * width + x - 1] == 1))
	  temp3[y * width + x] = 11;
	if ((temp3[(y - 1) * width + x] == 1) && (temp3[y * width + x + 1] == 1))
	  temp3[y * width + x] = 12;
	if ((temp3[(y + 1) * width + x] == 1) && (temp3[y * width + x + 1] == 1))
	  temp3[y * width + x] = 13;
	
	if ((temp3[(y - 1) * width + x + 1] == 1) && (temp3[(y - 1) * width + x] >= 2) && (temp3[y * width + x + 1] >= 2))
	  temp3[y * width + x] = 6;
	if ((temp3[(y + 1) * width + x + 1] == 1) && (temp3[(y + 1) * width + x] >= 2) && (temp3[y * width + x + 1] >= 2))
	  temp3[y * width + x] = 7;
	if ((temp3[(y + 1) * width + x - 1] == 1) && (temp3[(y + 1) * width + x] >= 2) && (temp3[y * width + x - 1] >= 2))
	  temp3[y * width + x] = 8;
	if ((temp3[(y - 1) * width + x - 1] == 1) && (temp3[(y - 1) * width + x] >= 2) && (temp3[y * width + x - 1] >= 2))
	  temp3[y * width + x] = 9;
      }
      if (temp3[y * width + x] <= 0)
      {
	if ((temp3[(y - 1) * width + x] >= 2))
	  temp3[y * width + x] = -3;
	if ((temp3[y * width + x + 1] >= 2))
	  temp3[y * width + x] = -4;
	if ((temp3[(y + 1) * width + x] >= 2))
	  temp3[y * width + x] = -5;
	if ((temp3[y * width + x - 1] >= 2))
	  temp3[y * width + x] = -2;
	
	if ((temp3[(y + 1) * width + x] >= 2) && (temp3[y * width + x - 1] >= 2))
	  temp3[y * width + x] = -12;
	if ((temp3[(y - 1) * width + x] >= 2) && (temp3[y * width + x - 1] >= 2))
	  temp3[y * width + x] = -13;
	if ((temp3[(y - 1) * width + x] >= 2) && (temp3[y * width + x + 1] >= 2))
	  temp3[y * width + x] = -10;
	if ((temp3[(y + 1) * width + x] >= 2) && (temp3[y * width + x + 1] >= 2))
	  temp3[y * width + x] = -11;
	
	if ((temp3[(y - 1) * width + x + 1] >= 2) && (temp3[(y - 1) * width + x] <= 0) && (temp3[y * width + x + 1] <= 0))
	  temp3[y * width + x] = -6;
	if ((temp3[(y + 1) * width + x + 1] >= 2) && (temp3[(y + 1) * width + x] <= 0) && (temp3[y * width + x + 1] <= 0))
	  temp3[y * width + x] = -7;
	if ((temp3[(y + 1) * width + x - 1] >= 2) && (temp3[(y + 1) * width + x] <= 0) && (temp3[y * width + x - 1] <= 0))
	  temp3[y * width + x] = -8;
	if ((temp3[(y - 1) * width + x - 1] >= 2) && (temp3[(y - 1) * width + x] <= 0) && (temp3[y * width + x - 1] <= 0))
	  temp3[y * width + x] = -9;
      }
    }
  }
  
  unsigned short kueste_1_2[] = {1011, 1012, 1013, 1014, 1015, 1016, 1017, 1018, 1019, 1020};
  unsigned short strand_1_2[] = {1205, 1214, 1215, 1216, 1217, 1218};
  
  for (y = 0; y < height; y++)
  {
    for (x = 0; x < width; x++)
    {
      daten[y * width + x].x_pos = x;
      daten[y * width + x].y_pos = y;
      daten[y * width + x].rot = 0;
      daten[y * width + x].ani = 0;
      daten[y * width + x].spieler = 7;
      daten[y * width + x].zufall = rand_(y * width + x) % 32;
      daten[y * width + x].spieler_ = 7;
      switch (temp3[y * width + x])
      {
	case 1:
	  if (perlin(x + 500, y + 100, 10, 10, 50) + perlin(x + 2000, y + 500, 1, 1, 40) > 45)
	    daten[y * width + x].bebauung = 1304 + (rand_(y * 70 + x) % 11) * 2;
	  else
	  {
	    int gras = perlin(x + 354, y + 975, 7, 7, 15) + perlin(x + 124, y + 45, 2, 2, 15);
	    if (gras >= 10)
	      daten[y * width + x].bebauung = 102;
	    else
	      daten[y * width + x].bebauung = 161 + gras;
	  }
	  break;
	case 2:
	  daten[y * width + x].bebauung = kueste_1_2[(x ^ y) % 10];//kueste_1_2[rand_(y * width + x) % 10];
	  daten[y * width + x].rot = 0;
	  break;
	case 3:
	  daten[y * width + x].bebauung = kueste_1_2[(x ^ y) % 10];//kueste_1_2[rand_(y * width + x) % 10];
	  daten[y * width + x].rot = 1;
	  break;
	case 4:
	  daten[y * width + x].bebauung = kueste_1_2[(x ^ y) % 10];//kueste_1_2[rand_(y * width + x) % 10];
	  daten[y * width + x].rot = 2;
	  break;
	case 5:
	  daten[y * width + x].bebauung = kueste_1_2[(x ^ y) % 10];//kueste_1_2[rand_(y * width + x) % 10];
	  daten[y * width + x].rot = 3;
	  break;
	case 6:
	  daten[y * width + x].bebauung = 1051 + rand_(y * width + x) % 3;
	  daten[y * width + x].rot = 0;
	  break;
	case 7:
	  daten[y * width + x].bebauung = 1051 + rand_(y * width + x) % 3;
	  daten[y * width + x].rot = 1;
	  break;
	case 8:
	  daten[y * width + x].bebauung = 1051 + rand_(y * width + x) % 3;
	  daten[y * width + x].rot = 2;
	  break;
	case 9:
	  daten[y * width + x].bebauung = 1051 + rand_(y * width + x) % 3;
	  daten[y * width + x].rot = 3;
	  break;
	case 10:
	  daten[y * width + x].bebauung = 1031 + rand_(y * width + x) % 4;
	  daten[y * width + x].rot = 0;
	  break;
	case 11:
	  daten[y * width + x].bebauung = 1031 + rand_(y * width + x) % 4;
	  daten[y * width + x].rot = 1;
	  break;
	case 12:
	  daten[y * width + x].bebauung = 1031 + rand_(y * width + x) % 4;
	  daten[y * width + x].rot = 2;
	  break;
	case 13:
	  daten[y * width + x].bebauung = 1031 + rand_(y * width + x) % 4;
	  daten[y * width + x].rot = 3;
	  break;
	case -2:
	  daten[y * width + x].bebauung = strand_1_2[(x ^ y) % 6];//strand_1_2[rand_(y * width + x) % 6];
	  daten[y * width + x].rot = 0;
	  break;
	case -3:
	  daten[y * width + x].bebauung = strand_1_2[(x ^ y) % 6];//strand_1_2[rand_(y * width + x) % 6];
	  daten[y * width + x].rot = 1;
	  break;
	case -4:
	  daten[y * width + x].bebauung = strand_1_2[(x ^ y) % 6];//strand_1_2[rand_(y * width + x) % 6];
	  daten[y * width + x].rot = 2;
	  break;
	case -5:
	  daten[y * width + x].bebauung = strand_1_2[(x ^ y) % 6];//strand_1_2[rand_(y * width + x) % 6];
	  daten[y * width + x].rot = 3;
	  break;
	case -6:
	  daten[y * width + x].bebauung = 1207;// + rand_(y * width + x) % 3;
	  daten[y * width + x].rot = 0;
	  break;
	case -7:
	  daten[y * width + x].bebauung = 1207;// + rand_(y * width + x) % 3;
	  daten[y * width + x].rot = 1;
	  break;
	case -8:
	  daten[y * width + x].bebauung = 1207;// + rand_(y * width + x) % 3;
	  daten[y * width + x].rot = 2;
	  break;
	case -9:
	  daten[y * width + x].bebauung = 1207;// + rand_(y * width + x) % 3;
	  daten[y * width + x].rot = 3;
	  break;
	case -10:
	  daten[y * width + x].bebauung = 1206;// + rand_(y * width + x) % 4;
	  daten[y * width + x].rot = 0;
	  break;
	case -11:
	  daten[y * width + x].bebauung = 1206;// + rand_(y * width + x) % 4;
	  daten[y * width + x].rot = 1;
	  break;
	case -12:
	  daten[y * width + x].bebauung = 1206;// + rand_(y * width + x) % 4;
	  daten[y * width + x].rot = 2;
	  break;
	case -13:
	  daten[y * width + x].bebauung = 1206;// + rand_(y * width + x) % 4;
	  daten[y * width + x].rot = 3;
	  break;
	case 20:
	  daten[y * width + x].bebauung = 1231 + rand_(y * width + x) % 4;
	  break;
	default:
	  if (temp[y * width + x] >= -10)
	    daten[y * width + x].bebauung = 1203;
	  else if (temp[y * width + x] >= -30)
	    daten[y * width + x].bebauung = 1204;
      //else if (temp[y * width + x] <= 20)
	  else daten[y * width + x].bebauung = 1201;
	  daten[y * width + x].ani = rand_(y * width + x) % 12;
	  break;
      }
    }
  }
  
  FILE *f = fopen("ausgabe.scp", "w");
  
  insel5_header_t header = {0, 0, 0, 116, 0, 70, 60, 0, 100, 100, 0, 0, {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}, 0, 0, 0, 1, {0,0,0,0,0,0,0,0,0,0,0,0,0,0}};
  fwrite(insel5_id, 1, 7, f);
  fwrite(&header, sizeof(insel5_header_t), 1, f);
  fwrite(inselhaus_id, 1, 10, f);
  inselhaus_header_t header2 = {0, 99, 70 * 60 * 8};
  fwrite(&header2, sizeof(inselhaus_header_t), 1, f);
  fwrite(daten, sizeof(inselfeld_t), 70 * 60, f);
  
  fclose(f);
}
